#!/usr/bin/env python3
# https://github.com/EONRaider/NETProtocols

__author__ = "EONRaider @ keybase.io/eonraider"

from ctypes import c_uint16, c_uint32

from netprotocols import Protocol


class TCP(Protocol):                # IETF RFC 793
    _fields_ = [
        ("sport", c_uint16),        # Source port
        ("dport", c_uint16),        # Destination port
        ("seq", c_uint32),          # Sequence number
        ("ack", c_uint32),          # Acknowledgement number
        ("offset", c_uint16, 4),    # Data offset
        ("reserved", c_uint16, 3),  # Reserved field
        ("flags", c_uint16, 9),     # TCP flag codes
        ("window", c_uint16),       # Window size
        ("chksum", c_uint16),       # TCP header checksum
        ("urg", c_uint16),          # Urgent pointer
    ]
    header_len = 32
    flag_names = "FIN", "SYN", "RST", "PSH", "ACK", "URG", "ECE", "CWR", "NS"

    def __init__(self, *,
                 sport: int,
                 dport: int,
                 seq: int,
                 ack: int,
                 offset: int,
                 reserved: int,
                 flags: int,
                 window: int,
                 chksum: int,
                 urg: int):
        super().__init__()
        self.sport = sport
        self.dport = dport
        self.seq = seq
        self.ack = ack
        self.offset = offset
        self.reserved = reserved
        self.flags = flags
        self.window = window
        self.chksum = chksum
        self.urg = urg

    @property
    def flags_hex(self) -> str:
        """
        Gets a string representation of the hexadecimal value of the
        TCP flags set on the segment.
        Ex: '0x018' or '0x010'
        """
        return format(self.flags, "#0{}x".format(5))

    @property
    def flags_txt(self) -> str:
        """
        Gets a space-separated string representation of the names of
        the TCP flags set on the segment.
        Ex: 'SYN ACK' or 'PSH ACK'
        """
        '''Yield the least-significant bit of a 9-bit flag value at each 
        iteration until there are no more bits left.'''
        flag_bits = ((self.flags >> shift) & 1 for shift in range(9))

        '''Yield the string representation of a flag name if the 
        corresponding bit is set.'''
        flags = (flag_name for flag_name, flag_bit in
                 zip(self.flag_names, flag_bits) if flag_bit == 1)
        return " ".join(flags)



#!/usr/bin/env python3
# https://github.com/EONRaider/NETProtocols

__author__ = "EONRaider @ keybase.io/eonraider"

from ctypes import c_ubyte, c_uint8, c_uint16

from netprotocols import Protocol


class ICMPv4(Protocol):         # IETF RFC 792
    _fields_ = [
        ("type", c_uint8),      # Control message type
        ("code", c_uint8),      # Control message subtype
        ("chksum", c_uint16),   # Header checksum
        ("_rest", c_ubyte * 4)  # Rest of header (contents vary)
    ]
    header_len = 8              # Length of the header in bytes
    icmpv4_types = {
        0: "Echo reply",
        3: "Destination Unreachable",
        4: "Source Quench",
        5: "Redirect Message",
        8: "Echo Request",
        9: "Router Advertisement",
        10: "Router Solicitation",
        11: "Time Exceeded",
        12: "Parameter Problem: Bad IP Header",
        13: "Timestamp",
        14: "Timestamp Reply",
        15: "Information Request",
        16: "Information Reply",
        17: "Address Mask Request",
        18: "Address Mask Reply",
        30: "Traceroute",
        42: "Extended Echo Request",
        43: "Extended Echo Reply"
    }

    def __init__(self, *,
                 type: int,
                 code: int,
                 chksum: int,
                 rest: bytes):
        super().__init__()
        self.type = type
        self.code = code
        self.chksum = chksum
        self.rest = (c_ubyte * 4)(*rest)

    @classmethod
    def decode(cls, packet: bytes):
        header = cls.from_buffer_copy(packet)
        header.rest = bytes(header._rest)
        return header

    @property
    def type_name(self) -> str:
        return self.icmpv4_types.get(
            self.type, "Unknown, Unassigned or Deprecated")


class ICMPv6(Protocol):           # IETF RFC 4443
    _fields_ = [
        ("type", c_uint8),        # Control message type
        ("code", c_uint8),        # Control message subtype
        ("chksum", c_uint16),     # Header checksum
        ("_m_body", c_ubyte * 4)  # Message body
    ]
    header_len = 8                # Length of the header in bytes
    icmpv6_types = {
        1: "Destination Unreachable",
        2: "Packet Too Big",
        3: "Time Exceeded",
        4: "Parameter Problem",
        100: "Private Experimentation",
        101: "Private Experimentation",
        127: "Reserved for Expansion of ICMPv6 Error Messages",
        128: "Echo Request",
        129: "Echo Reply",
        130: "Multicast Listener Query",
        131: "Multicast Listener Report",
        132: "Multicast Listener Done",
        133: "Router Solicitation",
        134: "Router Advertisement",
        135: "Neighbor Solicitation",
        136: "Neighbor Advertisement",
        137: "Redirect Message",
        138: "Router Renumbering",
        139: "ICMP Node Information Query",
        140: "ICMP Node Information Response",
        141: "Inverse Neighbor Discovery Solicitation Message",
        142: "Inverse Neighbor Discovery Advertisement Message",
        143: "Multicast Listener Discovery reports",
        144: "Home Agent Address Discovery Request Message",
        145: "Home Agent Address Discovery Reply Message",
        146: "Mobile Prefix Solicitation",
        147: "Mobile Prefix Advertisement",
        148: "Certification Path Solicitation",
        149: "Certification Path Advertisement",
        151: "Multicast Router Advertisement",
        152: "Multicast Router Solicitation",
        153: "Multicast Router Termination",
        155: "RPL Control Message",
        200: "Private Experimentation",
        201: "Private Experimentation",
        255: "Reserved for Expansion of ICMPv6 Informational Messages"
    }

    def __init__(self, *,
                 type: int,
                 code: int,
                 chksum: int,
                 m_body: bytes):
        super().__init__()
        self.type = type
        self.code = code
        self.chksum = chksum
        self.m_body = (c_ubyte * 4)(*m_body)

    @classmethod
    def decode(cls, packet: bytes):
        header = cls.from_buffer_copy(packet)
        header.m_body = bytes(header._m_body)
        return header

    @property
    def type_name(self) -> str:
        return self.icmpv6_types.get(
            self.type, "Unknown, Unassigned or Deprecated")



#!/usr/bin/env python3
# https://github.com/EONRaider/NETProtocols

__author__ = "EONRaider @ keybase.io/eonraider"

from netprotocols import Protocol


class Packet:
    def __init__(self, *protocols):
        for protocol in protocols:
            setattr(self, protocol.__class__.__name__, protocol)

    def __setattr__(self, protocol_name, protocol_class):
        valid_protocols = (cls.__name__ for cls in Protocol.__subclasses__())
        if protocol_name not in valid_protocols:
            raise AttributeError(f"Cannot build packet. Invalid protocol: "
                                 f"{protocol_name}")
        super().__setattr__(protocol_name.lower(), protocol_class)

    def __bytes__(self):
        return b"".join(proto for proto in self.__dict__.values())



#!/usr/bin/env python3
# https://github.com/EONRaider/NETProtocols

__author__ = "EONRaider @ keybase.io/eonraider"

import re
import socket
from ctypes import (
    Array,
    BigEndianStructure,
    create_string_buffer,
    c_ubyte,
    sizeof
)
from socket import inet_ntop, inet_pton, AF_INET
from typing import Union


class Protocol(BigEndianStructure):
    _pack_ = 1

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    def __init__(self, *args):
        super().__init__()

    def __str__(self):
        return create_string_buffer(sizeof(self))[:]

    @classmethod
    def decode(cls, packet: bytes):
        """Decode a raw network packet into a new instance of the
        protocol."""
        header = cls.from_buffer_copy(packet)
        return header

    @property
    def encapsulated_proto(self) -> Union[None, str]:
        """The string representation of the name of the encapsulated
        protocol. Override as required by the specific protocol being
        implemented."""
        return "undefined"

    @staticmethod
    def hdwr_to_addr_array(hdwr_addr: str) -> Array:
        """
        Convert an IEEE 802 MAC address to c_ubyte array of 6
        bytes.

        Ex: From "00:c0:ca:a8:19:74" to instance of Array with length
        equal to 6.
        """
        mac_to_bytes = b"".join(bytes.fromhex(octet)
                                for octet in re.split("[:-]", hdwr_addr))
        return (c_ubyte * 6)(*mac_to_bytes)

    @staticmethod
    def addr_array_to_hdwr(addr_array: Array) -> str:
        """
        Convert a c_ubyte array of 6 bytes to IEEE 802.3 MAC address.

        Ex: From instance of Array with length equal to 6 to
        "00:c0:ca:a8:19:74".
        """
        return ":".join(format(octet, "02x") for octet in bytes(addr_array))

    @staticmethod
    def proto_addr_to_array(proto_addr: str,
                            addr_family: socket.AddressFamily = AF_INET):
        """
        Convert an IPv4 address string in dotted-decimal notation
        to a c_ubyte array of 4 bytes or an IPv6 address string to a
        c_ubyte array of 16 bytes.

        Ex1: From "185.159.104.91" to instance of Array with length
        equal to 4.

        Ex2: From "fe80::200:86ff:fe05:80da" to instance of Array with
        length equal to 16.
        """
        try:
            addr_to_bytes = inet_pton(addr_family, proto_addr)
        except OSError:
            raise TypeError("Only addresses of family AF_INET and AF_INET6 are "
                            "supported.")
        return (c_ubyte * 4)(*addr_to_bytes) if addr_family == AF_INET else \
            (c_ubyte * 16)(*addr_to_bytes)

    @staticmethod
    def array_to_proto_addr(addr_array: Array,
                            addr_family: socket.AddressFamily = AF_INET) -> str:
        """
        Convert a packed IPv4/IPv6 address array to its RFC 791/2460
        string representation.

        Ex1: From instance of Array with length equal to 4 to
        "185.159.104.91"

        Ex2: From instance of Array with length equal to 16 to
        "fe80::200:86ff:fe05:80da"
        """
        try:
            return inet_ntop(addr_family, bytes(addr_array))
        except OSError:
            raise TypeError("Only addresses of family AF_INET and AF_INET6 are "
                            "supported.")



#!/usr/bin/env python3
# https://github.com/EONRaider/NETProtocols

__author__ = "EONRaider @ keybase.io/eonraider"

from ctypes import c_uint16

from netprotocols import Protocol


class UDP(Protocol):          # IETF RFC 768
    _fields_ = [
        ("sport", c_uint16),  # Source port
        ("dport", c_uint16),  # Destination port
        ("len", c_uint16),    # Header length
        ("chksum", c_uint16)  # Header checksum
    ]
    header_len = 8

    def __init__(self, *,
                 sport: int,
                 dport: int,
                 len: int,
                 chksum: int):
        super().__init__()
        self.sport = sport
        self.dport = dport
        self.len = len
        self.chksum = chksum



#!/usr/bin/env python3
# https://github.com/EONRaider/NETProtocols

__author__ = "EONRaider @ keybase.io/eonraider"

from ctypes import c_ubyte, c_uint8, c_uint16, c_uint32
from socket import AF_INET6

from netprotocols import Protocol


class IP:
    protocol_numbers = {  # As defined by RFC 790
        0x01: "ICMP",
        0x02: "IGMP",
        0x06: "TCP",
        0x11: "UDP",
        0x3a: "IPv6-ICMP"
    }


class IPv4(IP, Protocol):          # IETF RFC 791
    _fields_ = [
        ("version", c_uint8, 4),   # Protocol version
        ("ihl", c_uint8, 4),       # Internet header length
        ("dscp", c_uint8, 6),      # Differentiated services code point
        ("ecp", c_uint8, 2),       # Explicit congestion notification
        ("len", c_uint16),         # Total packet length
        ("id", c_uint16),          # Identification
        ("flags", c_uint16, 3),    # Fragmentation control flags
        ("offset", c_uint16, 13),  # Fragment offset
        ("ttl", c_uint8),          # Time to live
        ("proto", c_uint8),        # Encapsulated protocol
        ("chksum", c_uint16),      # Header checksum
        ("_src", c_ubyte * 4),     # Source address
        ("_dst", c_ubyte * 4)      # Destination address
    ]
    header_len = 20                # Length of the header in bytes

    def __init__(self, *,
                 version: int,
                 ihl: int,
                 dscp: int,
                 ecp: int,
                 len: int,
                 id: int,
                 flags: int,
                 offset: int,
                 ttl: int,
                 proto: int,
                 chksum: int,
                 src: str,
                 dst: str):
        super().__init__()
        self.version = version
        self.ihl = ihl
        self.dscp = dscp
        self.ecp = ecp
        self.len = len
        self.id = id
        self.flags = flags
        self.offset = offset
        self.ttl = ttl
        self.proto = proto
        self.chksum = chksum
        self.src = self.proto_addr_to_array(src)
        self.dst = self.proto_addr_to_array(dst)

    @classmethod
    def decode(cls, packet: bytes):
        header = cls.from_buffer_copy(packet)
        header.src = cls.array_to_proto_addr(header._src)
        header.dst = cls.array_to_proto_addr(header._dst)
        return header

    @property
    def encapsulated_proto(self):
        return self.protocol_numbers.get(self.proto, None)


class IPv6(IP, Protocol):           # IETF RFC 2460 / 8200
    _fields_ = [
        ("version", c_uint32, 4),   # Protocol version
        ("tclass", c_uint32, 8),    # Traffic class
        ("flabel", c_uint32, 20),   # Flow label
        ("payload_len", c_uint16),  # Payload length
        ("next_header", c_uint8),   # Type of next header
        ("hop_limit", c_uint8),     # Hop limit (replaces IPv4 TTL)
        ("_src", c_ubyte * 16),     # Source address
        ("_dst", c_ubyte * 16)      # Destination address
    ]
    header_len = 40                 # Length of the header in bytes

    def __init__(self, *,
                 version: int,
                 tclass: int,
                 flabel: int,
                 payload_len: int,
                 next_header: int,
                 hop_limit: int,
                 src: str,
                 dst: str):
        super().__init__()
        self.version = version
        self.tclass = tclass
        self.flabel = flabel
        self.payload_len = payload_len
        self.next_header = next_header
        self.hop_limit = hop_limit
        self.src = self.proto_addr_to_array(src, addr_family=AF_INET6)
        self.dst = self.proto_addr_to_array(dst, addr_family=AF_INET6)

    @classmethod
    def decode(cls, packet: bytes):
        header = cls.from_buffer_copy(packet)
        header.src = cls.array_to_proto_addr(header._src, addr_family=AF_INET6)
        header.dst = cls.array_to_proto_addr(header._dst, addr_family=AF_INET6)
        return header

    @property
    def encapsulated_proto(self):
        return self.protocol_numbers.get(self.next_header, None)



from netprotocols.base.protocol import Protocol
from netprotocols.base.packet import Packet
from netprotocols.layer2.ethernet import Ethernet
from netprotocols.layer2.arp import ARP
from netprotocols.layer3.icmp import ICMPv4, ICMPv6
from netprotocols.layer3.ip import IPv4, IPv6
from netprotocols.layer4.tcp import TCP
from netprotocols.layer4.udp import UDP



#!/usr/bin/env python3
# https://github.com/EONRaider/NETProtocols

__author__ = "EONRaider @ keybase.io/eonraider"

from ctypes import c_ubyte, c_uint16

from netprotocols import Protocol


class Ethernet(Protocol):       # IEEE 802.3 standard
    _fields_ = [
        ("_dst", c_ubyte * 6),  # Destination hardware address
        ("_src", c_ubyte * 6),  # Source hardware address
        ("eth", c_uint16)       # Ethertype
    ]
    header_len = 14             # Length of the header in bytes
    ethertypes = {
        0x0806: "ARP",
        0x0800: "IPv4",
        0x86dd: "IPv6"
    }

    def __init__(self, *, dst: str, src: str, eth: int):
        super().__init__()
        self.dst = self.hdwr_to_addr_array(dst)
        self.src = self.hdwr_to_addr_array(src)
        self.eth = eth

    @classmethod
    def decode(cls, packet: bytes):
        header = cls.from_buffer_copy(packet)
        header.src = cls.addr_array_to_hdwr(header._src)
        header.dst = cls.addr_array_to_hdwr(header._dst)
        return header

    @property
    def encapsulated_proto(self):
        return self.ethertypes.get(self.eth, None)



#!/usr/bin/env python3
# https://github.com/EONRaider/NETProtocols

__author__ = 'EONRaider @ keybase.io/eonraider'


from ctypes import c_ubyte, c_uint8, c_uint16

from netprotocols import Protocol


class ARP(Protocol):            # IETF RFC 826
    _fields_ = [
        ("htype", c_uint16),    # Hardware type
        ("ptype", c_uint16),    # Protocol type
        ("hlen", c_uint8),      # Hardware length
        ("plen", c_uint8),      # Protocol length
        ("oper", c_uint16),     # Operation code
        ("_sha", c_ubyte * 6),  # Sender hardware address
        ("_spa", c_ubyte * 4),  # Sender protocol address
        ("_tha", c_ubyte * 6),  # Target hardware address
        ("_tpa", c_ubyte * 4),  # Target protocol address
    ]
    header_len = 28             # Length of the header in bytes

    def __init__(self, *,
                 htype: int,
                 ptype: int,
                 hlen: int,
                 plen: int,
                 oper: int,
                 sha: str,
                 spa: str,
                 tha: str,
                 tpa: str):
        super().__init__()
        self.htype = htype
        self.ptype = ptype
        self.hlen = hlen
        self.plen = plen
        self.oper = oper
        self.sha = self.hdwr_to_addr_array(sha)
        self.spa = self.proto_addr_to_array(spa)
        self.tha = self.hdwr_to_addr_array(tha)
        self.tpa = self.proto_addr_to_array(tpa)

    @classmethod
    def decode(cls, packet: bytes):
        header = cls.from_buffer_copy(packet)
        header.sha = cls.addr_array_to_hdwr(header._sha)
        header.spa = cls.array_to_proto_addr(header._spa)
        header.tha = cls.addr_array_to_hdwr(header._tha)
        header.tpa = cls.array_to_proto_addr(header._tpa)
        return header
